//*******************************************************************************
// file  v4l2cam2dev.cpp
//
// attention
// Copyright (C) 2020 MACNICA,Inc. All Rights Reserved.\n
//   This software is licensed "AS IS".
//   Please perform use of this software by a user's own responsibility and expense.
//   It cannot guarantee in the maker side about the damage which occurred by the ab-
//   ility not to use or use this software, and all damage that occurred secondarily.
//*******************************************************************************

#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>
#include <linux/videodev2.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <signal.h>
#include <fstream>
#include "logger.h"
#include "V4l2Device.h"
#include "V4l2Capture.h"
#include "V4l2Output.h"

int stop=0;

#define VFB_ADDR                0x7F000000
#define VFB_SIZE                0x00800000
#define LW_BRIDGE_ADDR          0x0
#define LW_BRIDGE_SIZE          0x1000

/* VFB II control registers */
#define VFB_OFFSET              0x000
#define ALTVIPFB2_CONTROL       0
#define ALTVIPFB2_STATUS        1
#define ALTVIPFB2_INTERRUPT     2
#define ALTVIPFB2_FRAME_COUNTER 3
#define ALTVIPFB2_FRAME_DROP    4
#define ALTVIPFB2_FRAME_INFO    5
#define ALTVIPFB2_FRAME_START   6
#define ALTVIPFB2_FRAME_READER  7
#define ALTVIPFB2_FRAME_MISC    8
#define ALTVIPFB2_FRAME_LOCK    9
#define ALTVIPFB2_FRAME_IRATE   10
#define ALTVIPFB2_FRAME_ORATE   11

/* PIO registers */
#define PIO_OFFSET              0x040
#define ALTPIO_DATA             0

/* Scaler II control registers */
#define SCL_OFFSET              0x200
#define ALTVIPSCL2_CONTROL      0
#define ALTVIPSCL2_STATUS       1
#define ALTVIPSCL2_INTERRUPT    2
#define ALTVIPSCL2_WIDTH        3
#define ALTVIPSCL2_HEIGHT       4

void    serdes_init(void);
void    ov5642_init(void);

/* ---------------------------------------------------------------------------
**  SIGINT handler
** -------------------------------------------------------------------------*/
void sighandler(int)
{ 
    printf("SIGINT\n");
    stop = 1;
}

/* ---------------------------------------------------------------------------
**  main
** -------------------------------------------------------------------------*/
int main(int argc, char* argv[]) 
{   
    int             result = 0;
    int             verbose = 0;
    const char      *out_devname = "/dev/video0";   
    V4l2IoType ioTypeOut = IOTYPE_MMAP;
    int             width = 640;
    int             height = 480;
    int             fps = 30;
    
    int             fd_vfb = 0;         // uio0 file discripter
    unsigned char   *vfb_mem = NULL;    // uio0 mmap pointer
    int             fd_reg = 0;         // uio1 file discripter
    unsigned char   *reg_mem = NULL;    // uio1 mmap pointer
    unsigned long   *vfb_reg = NULL;    // Frame buffer register base pointer
    unsigned long   *scl_reg = NULL;    // Scaler register base pointer

    int c = 0;
    while ((c = getopt (argc, argv, "hP:F:v::w" "W:H:F:Z:")) != -1)
    {
        switch (c)
        {
            case 'v':   verbose   = 1; if (optarg && *optarg=='v') verbose++;  break;
            case 'w':   ioTypeOut = IOTYPE_READWRITE; break;
            
            case 'W':   width = atoi(optarg); break;
            case 'H':   height = atoi(optarg); break;
            case 'F':   fps = atoi(optarg); break;
            
            case 'h':
            {
                std::cout << argv[0] << " [-v[v]] [-W width] [-H height] source_device dest_device" << std::endl;
                std::cout << "\t -v            : verbose " << std::endl;
                std::cout << "\t -vv           : very verbose " << std::endl;
                std::cout << "\t -W width      : V4L2 capture width (default "<< width << ")" << std::endl;
                std::cout << "\t -H height     : V4L2 capture height (default "<< height << ")" << std::endl;
                std::cout << "\t -F fps        : V4L2 capture framerate (default "<< fps << ")" << std::endl;
                std::cout << "\t -w            : V4L2 capture using write interface (default use memory mapped buffers)" << std::endl;
                std::cout << "\t dest_device   : V4L2 capture device (default "<< out_devname << ")" << std::endl;
                exit(0);
            }
        }
    }
    if (optind<argc)
    {
        out_devname = argv[optind];
        optind++;
    }

    try 
    {
        // initialize log4cpp
        initLogger(verbose);

        if ((fd_vfb = open("/dev/uio0", O_RDWR)) != -1) 
        {
            vfb_mem = (unsigned char *)mmap(NULL, VFB_SIZE, PROT_READ | PROT_WRITE, MAP_SHARED, fd_vfb, 0);
            if(vfb_mem == (unsigned char *)-1)
            {
                throw std::logic_error("/dev/uio0 mmap error!");
            }
        } else
        {
            throw std::logic_error("/dev/uio0 open error!");
        }

        if ((fd_reg = open("/dev/uio1", O_RDWR)) != -1) 
        {
            reg_mem = (unsigned char *)mmap(NULL, LW_BRIDGE_SIZE, PROT_READ | PROT_WRITE, MAP_SHARED, fd_reg, LW_BRIDGE_ADDR);
            if(reg_mem == (unsigned char *)-1)
            {
                throw std::logic_error("/dev/uio1 mmap error!");
            }
        } else
        {
            throw std::logic_error("/dev/uio0 open error!");
        }
        // Initialize SerDes
        serdes_init();
        // Reset CAM_PWDN
        unsigned long *pio_reg = (unsigned long*)(reg_mem + PIO_OFFSET);
        pio_reg[ALTPIO_DATA] = 1;
        usleep(10000);
        pio_reg[ALTPIO_DATA] = 0;
        // Initialize OV5642
        ov5642_init();

        vfb_reg = (unsigned long*)(reg_mem + VFB_OFFSET);
        scl_reg = (unsigned long*)(reg_mem + SCL_OFFSET);
        // Initialize Scaler
        scl_reg[ALTVIPSCL2_WIDTH] = width;
        scl_reg[ALTVIPSCL2_HEIGHT] = height;
        scl_reg[ALTVIPSCL2_CONTROL] = 0x01; // Scaler Start

        // Initialize Frame Buffer
        vfb_reg[ALTVIPFB2_CONTROL] = 0x01;  // Frame buffer Start
    
        // init V4L2 output interface
        V4L2DeviceParameters outparam(out_devname, V4L2_PIX_FMT_BGR24, width, height, fps, ioTypeOut, verbose);
        V4l2Output* videoOutput = V4l2Output::create(outparam);
        if (videoOutput == NULL)
        {   
            LOG(WARN) << "Cannot create V4L2 output interface for device:" << out_devname; 
        } else
        {       
            LOG(NOTICE) << "Start generating frames to " << out_devname; 
            signal(SIGINT,sighandler);
            int picture_size = videoOutput->getBufferSize();
            
            while (!stop) 
            {
                if(vfb_reg[ALTVIPFB2_FRAME_INFO] & 0x80000000)
                {
                    unsigned char *frame_adr = vfb_mem + (vfb_reg[ALTVIPFB2_FRAME_START] - VFB_ADDR);
                    int wsize = videoOutput->write((char*)frame_adr, picture_size);
                    LOG(DEBUG) << "Copied " << picture_size << " " << wsize; 
                }
                vfb_reg[ALTVIPFB2_FRAME_MISC] = 0x1;
            }
            delete videoOutput;
        }
    }
    catch (const std::exception& error) 
    {
        std::cerr << "[ ERROR ] " << error.what() << std::endl;
        result = 1;
    }
    catch (...) 
    {
        std::cerr << "[ ERROR ] Unknown/internal exception happened." << std::endl;
        result = 1;
    }
    if(scl_reg != NULL) scl_reg[ALTVIPSCL2_CONTROL] = 0x00; // Scaler Stop
    if(vfb_reg != NULL) vfb_reg[ALTVIPFB2_CONTROL] = 0x00;  // Frame Buffer Stop

    if(vfb_mem != NULL) munmap(vfb_mem, VFB_SIZE);
    if(reg_mem != NULL) munmap(reg_mem, LW_BRIDGE_SIZE);

    if(fd_vfb != 0) close(fd_vfb);
    if(fd_reg != 0) close(fd_reg);

    return result;
}
